/* 
 * Copyright (c) 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995, 1996
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Format and print AppleTalk packets.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sys/param.h>
#include <sys/socket.h>

#include "interfac.h"
#include "a2name.h"
#include "ethertyp.h"
#include "extract.h"                /* must come after interface.h */
#include "appletal.h"


static struct tok type2str[] = {
       { ddpRTMP,        "rtmp"    },
       { ddpRTMPrequest, "rtmpReq" },
       { ddpECHO,        "echo"    },
       { ddpIP,          "IP"      },
       { ddpARP,         "ARP"     },
       { ddpKLAP,        "KLAP"    },
       { 0,              NULL      }
     };

struct aarp {
       u_short htype, ptype;
       u_char  halen, palen;
       u_short op;
       u_char  hsaddr[6];
       u_char  psaddr[4];
       u_char  hdaddr[6];
       u_char  pdaddr[4];
     };

static char tstr[] = "[|atalk]";

static void atp_print (const struct atATP *, u_int);
static void atp_bitmap_print (u_char);
static void nbp_print (const struct atNBP *, u_int, u_short, u_char, u_char);
static const char *print_cstring (const char *, const u_char *);
static const struct atNBPtuple *nbp_tuple_print (const struct atNBPtuple *,
                                                 const u_char *,
                                                 u_short, u_char, u_char);
static const struct atNBPtuple *nbp_name_print (const struct atNBPtuple *,
                                                const u_char *);
static const char *ataddr_string (u_short, u_char);
static void ddp_print (const u_char *, u_int, int, u_short, u_char, u_char);
static const char *ddpskt_string (int);

/*
 * Print AppleTalk LLAP packets.
 */
void llap_print (const u_char * bp, u_int length)
{
  const struct LAP *lp;
  const struct atDDP *dp;
  const struct atShortDDP *sdp;
  u_short snet;

  lp = (struct LAP *) bp;
  bp     += sizeof(*lp);
  length -= sizeof(*lp);
  switch (lp->type)
  { 
    case lapShortDDP:
         if (length < ddpSSize)
         {
           PRINTF (" [|sddp %d]", length);
           return;
         }
         sdp = (const struct atShortDDP *) bp;
         PRINTF ("%s.%s", ataddr_string (0, lp->src),
                 ddpskt_string (sdp->srcSkt));
         PRINTF (" > %s.%s:", ataddr_string (0, lp->dst),
                 ddpskt_string (sdp->dstSkt));
         bp     += ddpSSize;
         length -= ddpSSize;
         ddp_print (bp, length, sdp->type, 0, lp->src, sdp->srcSkt);
         break;

    case lapDDP:
         if (length < ddpSize)
         {
           PRINTF (" [|ddp %d]", length);
           return;
         }
         dp = (const struct atDDP *) bp;
         snet = EXTRACT_16BITS (&dp->srcNet);
         PRINTF ("%s.%s", ataddr_string (snet, dp->srcNode),
                 ddpskt_string (dp->srcSkt));
         PRINTF (" > %s.%s:",
                 ataddr_string (EXTRACT_16BITS (&dp->dstNet), dp->dstNode),
                 ddpskt_string (dp->dstSkt));
         bp     += ddpSize;
         length -= ddpSize;
         ddp_print (bp, length, dp->type, snet, dp->srcNode, dp->srcSkt);
         break;

#ifdef notdef
    case lapKLAP:
         klap_print (bp, length);
         break;
#endif

    default:
         PRINTF ("%d > %d at-lap#%d %d", lp->src, lp->dst, lp->type, length);
         break;
  }
}

/*
 * Print EtherTalk/TokenTalk packets (or FDDITalk, or whatever it's called
 * when it runs over FDDI; yes, I've seen FDDI captures with AppleTalk
 * packets in them).
 */
void atalk_print (const u_char * bp, u_int length)
{
  const struct LAP        *lp;
  const struct atDDP      *dp;
  const struct atShortDDP *sdp;
  u_short snet;

  lp  = (struct LAP *) bp;
  bp += sizeof(*lp);
  length -= sizeof(*lp);

  switch (lp->type)
  {
    case lapShortDDP:
         if (length < ddpSSize)
         {
           PRINTF (" [|sddp %d]", length);
           return;
         }
         sdp = (const struct atShortDDP *) bp;
         PRINTF ("%s.%s",
                 ataddr_string (0, lp->src), ddpskt_string (sdp->srcSkt));
         PRINTF (" > %s.%s:",
                 ataddr_string (0, lp->dst), ddpskt_string (sdp->dstSkt));
         bp     += ddpSSize;
         length -= ddpSSize;
         ddp_print (bp, length, sdp->type, 0, lp->src, sdp->srcSkt);
         break;

    case lapDDP:
         if (length < ddpSize)
         {
           PRINTF (" [|ddp %d]", length);
           return;
         }
         dp = (const struct atDDP *) bp;
         snet = EXTRACT_16BITS (&dp->srcNet);
         PRINTF ("%s.%s", ataddr_string (snet, dp->srcNode),
                 ddpskt_string (dp->srcSkt));
         PRINTF (" > %s.%s:",
                 ataddr_string (EXTRACT_16BITS (&dp->dstNet), dp->dstNode),
                 ddpskt_string (dp->dstSkt));
         bp += ddpSize;
         length -= ddpSize;
         ddp_print (bp, length, dp->type, snet, dp->srcNode, dp->srcSkt);
         break;

#ifdef notdef
    case lapKLAP:
         klap_print (bp, length);
         break;
#endif

    default:
         PRINTF ("%d > %d at-lap#%d %d",
                 lp->src, lp->dst, lp->type, length);
         break;
     }
}

/*
 * XXX should probably pass in the snap header and do checks like arp_print()
 */
void aarp_print (const u_char * bp, u_int length)
{
  const struct aarp *ap;

#define AT(member) ataddr_string((ap->member[1]<<8)|ap->member[2],ap->member[3])

  PUTS ("aarp ");
  ap = (const struct aarp *) bp;
  if (ap->htype == 1 && ap->ptype == ETHERTYPE_ATALK &&
      ap->halen == 6 && ap->palen == 4)
    switch (ap->op)
    {
      case 1:                /* request */
           PRINTF ("who-has %s tell %s", AT(pdaddr), AT(psaddr));
           return;

      case 2:                /* response */
           PRINTF ("reply %s is-at %s", AT(psaddr), etheraddr_string(ap->hsaddr));
           return;

      case 3:                /* probe (oy!) */
           PRINTF ("probe %s tell %s", AT(pdaddr), AT(psaddr));
           return;
    }
  PRINTF ("len %d op %d htype %d ptype %#x halen %d palen %d",
          length, ap->op, ap->htype, ap->ptype, ap->halen, ap->palen);
}

static void ddp_print (const u_char * bp, u_int length, int t,
                       u_short snet, u_char snode, u_char skt)
{
  switch (t)
  {
    case ddpNBP:
         nbp_print ((const struct atNBP *) bp, length, snet, snode, skt);
         break;

    case ddpATP:
         atp_print ((const struct atATP *) bp, length);
         break;

    default:
         PRINTF (" at-%s %d", tok2str (type2str, "#%d", t), length);
         break;
  }
}

static void atp_print (const struct atATP *ap, u_int length)
{
  char      c;
  u_int32_t data;

  if ((const u_char*)(ap+1) > snapend)
  {
    /* Just bail if we don't have the whole chunk. */
    PUTS (tstr);
    return;
  }
  length -= sizeof(*ap);
  switch (ap->control & 0xc0)
  {
    case atpReqCode:
         PRINTF (" atp-req%s %d",
                 ap->control & atpXO ? " " : "*",
                 EXTRACT_16BITS (&ap->transID));

         atp_bitmap_print (ap->bitmap);

         if (length != 0)
            PRINTF (" [len=%d]", length);

         switch (ap->control & (atpEOM | atpSTS))
         {
           case atpEOM:
                PUTS (" [EOM]");
                break;
           case atpSTS:
                PUTS (" [STS]");
                break;
           case atpEOM | atpSTS:
                PUTS (" [EOM,STS]");
                break;
         }
         break;

    case atpRspCode:
         PRINTF (" atp-resp%s%d:%d (%d)",
                 ap->control & atpEOM ? "*" : " ",
                 EXTRACT_16BITS (&ap->transID), ap->bitmap, length);
         switch (ap->control & (atpXO | atpSTS))
         {
           case atpXO:
                PUTS (" [XO]");
                break;
           case atpSTS:
                PUTS (" [STS]");
                break;
           case atpXO | atpSTS:
                PUTS (" [XO,STS]");
                break;
         }
         break;

    case atpRelCode:
         PRINTF (" atp-rel  %d", EXTRACT_16BITS (&ap->transID));

         atp_bitmap_print (ap->bitmap);

         /* length should be zero */
         if (length)
            PRINTF (" [len=%d]", length);

         /* there shouldn't be any control flags */
         if (ap->control & (atpXO | atpEOM | atpSTS))
         {
           c = '[';
           if (ap->control & atpXO)
           {
             PRINTF ("%cXO", c);
             c = ',';
           }
           if (ap->control & atpEOM)
           {
             PRINTF ("%cEOM", c);
             c = ',';
           }
           if (ap->control & atpSTS)
           {
             PRINTF ("%cSTS", c);
             c = ',';
           }
           PUTCHAR (']');
         }
         break;

    default:
         PRINTF (" atp-0x%x  %d (%d)", ap->control,
                 EXTRACT_16BITS (&ap->transID), length);
         break;
  }

  data = EXTRACT_32BITS (&ap->userData);
  if (data)
     PRINTF (" 0x%lx", data);
}

static void atp_bitmap_print (u_char bm)
{
  char c;
  int  i;

  /*
   * The '& 0xff' below is needed for compilers that want to sign
   * extend a u_char, which is the case with the Ultrix compiler.
   * (gcc is smart enough to eliminate it, at least on the Sparc).
   */
  if ((bm + 1) & (bm & 0xff))
  {
    c = '<';
    for (i = 0; bm; ++i)
    {
      if (bm & 1)
      {
        PRINTF ("%c%d", c, i);
        c = ',';
      }
      bm >>= 1;
    }
    PUTCHAR ('>');
  }
  else
  {
    for (i = 0; bm; ++i)
      bm >>= 1;
    if (i > 1)
         PRINTF ("<0-%d>", i - 1);
    else PUTS ("<0>");
  }
}

static void nbp_print (const struct atNBP *np, u_int length, u_short snet,
                       u_char snode, u_char skt)
{
  const struct atNBPtuple *tp = (struct atNBPtuple*) ((u_char*)np + nbpHeaderSize);
  const u_char *ep;
  int   i;

  length -= nbpHeaderSize;
  if (length < 8)
  {
    /* must be room for at least one tuple */
    PRINTF (" truncated-nbp %d", length + nbpHeaderSize);
    return;
  }
  /* ep points to end of available data */
  ep = snapend;
  if ((const u_char *) tp > ep)
  {
    PUTS (tstr);
    return;
  }
  switch (i = np->control & 0xf0)
  {
    case nbpBrRq:
    case nbpLkUp:
         PRINTF (i == nbpLkUp ? " nbp-lkup %d:" : " nbp-brRq %d:", np->id);
         if ((const u_char *) (tp + 1) > ep)
         {
           PUTS (tstr);
           return;
         }
         nbp_name_print (tp, ep);
         /*
          * look for anomalies: the spec says there can only
          * be one tuple, the address must match the source
          * address and the enumerator should be zero.
          */
         if ((np->control & 0xf) != 1)
            PRINTF (" [ntup=%d]", np->control & 0xf);
         if (tp->enumerator)
            PRINTF (" [enum=%d]", tp->enumerator);
         if (EXTRACT_16BITS (&tp->net) != snet ||
             tp->node != snode || tp->skt != skt)
            PRINTF (" [addr=%s.%d]",
                    ataddr_string(EXTRACT_16BITS(&tp->net),tp->node),tp->skt);
         break;

    case nbpLkUpReply:
         PRINTF (" nbp-reply %d:", np->id);

         /* print each of the tuples in the reply */
         for (i = np->control & 0xf; --i >= 0 && tp;)
             tp = nbp_tuple_print (tp, ep, snet, snode, skt);
         break;

    default:
         PRINTF (" nbp-0x%x  %d (%d)", np->control, np->id, length);
         break;
     }
}

/*
 * print a counted string
 */
static const char * print_cstring (const char *cp, const u_char * ep)
{
  u_int length;

  if (cp >= (const char *) ep)
  {
    PUTS (tstr);
    return (0);
  }
  length = *cp++;

  /* Spec says string can be at most 32 bytes long */
  if (length > 32)
  {
    PRINTF ("[len=%u]", length);
    return (0);
  }
  while ((int)--length >= 0)
  {
    if (cp >= (char *) ep)
    {
      PUTS (tstr);
      return (0);
    }
    PUTCHAR (*cp++);
  }
  return (cp);
}

static const struct atNBPtuple * nbp_tuple_print (
                    const struct atNBPtuple *tp, const u_char * ep,
                    u_short snet, u_char snode, u_char skt)
{
  const struct atNBPtuple *tpn;

  if ((const u_char *) (tp + 1) > ep)
  {
    PUTS (tstr);
    return 0;
  }
  tpn = nbp_name_print (tp, ep);

  /* if the enumerator isn't 1, print it */
  if (tp->enumerator != 1)
     PRINTF ("(%d)", tp->enumerator);

  /* if the socket doesn't match the src socket, print it */
  if (tp->skt != skt)
     PRINTF (" %d", tp->skt);

  /* if the address doesn't match the src address, it's an anomaly */
  if (EXTRACT_16BITS (&tp->net) != snet || tp->node != snode)
     PRINTF (" [addr=%s]",
             ataddr_string(EXTRACT_16BITS(&tp->net),tp->node));

  return (tpn);
}

static const struct atNBPtuple *nbp_name_print (
                    const struct atNBPtuple *tp, const u_char * ep)
{
  const char *cp = (const char *) tp + nbpTupleSize;

  PUTCHAR (' ');

  /* Object */
  PUTCHAR ('"');
  cp = print_cstring (cp, ep);
  if (cp)
  {
    /* Type */
    PUTCHAR (':');
    cp = print_cstring (cp, ep);
    if (cp)
    {
      /* Zone */
      PUTCHAR ('@');
      cp = print_cstring (cp, ep);
      if (cp)
         PUTCHAR ('"');
    }
  }
  return ((const struct atNBPtuple*) cp);
}


#define HASHNAMESIZE 4096

static struct hnamemem hnametable[HASHNAMESIZE];

static const char *ataddr_string (u_short atnet, u_char athost)
{
  struct hnamemem *tp, *tp2;
  int    i = (atnet << 8) | athost;
  char   nambuf[256];
  static int first = 1;
  FILE  *fp;

  /*
   * if this is the first call, see if there's an AppleTalk
   * number to name map file.
   */

  if (first && (first = 0, !nflag) &&
      ((fp = fopen(atalk_file, "rt"))) != NULL)
  {
    char line[256];
    int  i1, i2, i3;

    while (fgets(line,sizeof(line),fp))
    {
      if (line[0] == '\n' || line[0] == 0 || line[0] == '#')
         continue;
      if (sscanf(line,"%d.%d.%d %s", &i1, &i2, &i3, nambuf) == 4)
      {
        /* got a hostname. */
        i3 |= ((i1 << 8) | i2) << 8;
      }
      else if (sscanf (line, "%d.%d %s", &i1, &i2, nambuf) == 3)
      {
        /* got a net name */
        i3 = (((i1 << 8) | i2) << 8) | 255;
      }
      else
        continue;

      for (tp = &hnametable[i3 & (HASHNAMESIZE - 1)]; tp->nxt; tp = tp->nxt)
          ;
      tp->addr = i3;
      tp->nxt  = newhnamemem();
      tp->name = savestr (nambuf);
    }
    fclose (fp);
  }

  for (tp = &hnametable[i & (HASHNAMESIZE - 1)]; tp->nxt; tp = tp->nxt)
      if (tp->addr == i)
         return (tp->name);

  /* didn't have the node name -- see if we've got the net name */
  i |= 255;
  for (tp2 = &hnametable[i & (HASHNAMESIZE - 1)]; tp2->nxt; tp2 = tp2->nxt)
    if (tp2->addr == i)
    {
      tp->addr = (atnet << 8) | athost;
      tp->nxt  = newhnamemem ();
      sprintf (nambuf, "%s.%d", tp2->name, athost);
      tp->name = savestr (nambuf);
      return (tp->name);
    }

  tp->addr = (atnet << 8) | athost;
  tp->nxt = newhnamemem ();
  if (athost != 255)
       sprintf (nambuf, "%d.%d.%d", atnet >> 8, atnet & 0xff, athost);
  else sprintf (nambuf, "%d.%d", atnet >> 8, atnet & 0xff);
  tp->name = savestr (nambuf);

  return (tp->name);
}

static struct tok skt2str[] = {
       { rtmpSkt, "rtmp" },      /* routing table maintenance */
       { nbpSkt,  "nis"  },      /* name info socket */
       { echoSkt, "echo" },      /* AppleTalk echo protocol */
       { zipSkt,  "zip"  },      /* zone info protocol */
       { 0, NULL}
     };

static const char *ddpskt_string (int skt)
{
  static char buf[8];

  if (nflag)
  {
    sprintf (buf, "%d", skt);
    return (buf);
  }
  return tok2str (skt2str, "%d", skt);
}
