/*
 *
 *   BBN Technologies
 *   10 Moulton St.
 *   Cambridge, MA 02138
 *   (617) 873-2000
 *
 *   Copyright (C) 1998, 1999
 *   This software is subject to copyright protection under the laws of
 *   the United States and other countries.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that: (1) source code
 *  distributions retain the above copyright notice and this paragraph
 *  in its entirety, (2) distributions including binary code include
 *  the above copyright notice and this paragraph in its entirety in
 *  the documentation or other materials provided with the
 *  distribution, and (3) all advertising materials mentioning features
 *  or use of this software display the following acknowledgement:
 *  ``This product includes software developed by the BBN Technologies
 *  and its contributors.'' Neither the name of t BBN nor the names of
 *  its contributors may be used to endorse or promote products derived
 *  from this software without specific prior written permission.  THIS
 *  SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/param.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/tcpip.h>

#include "interfac.h"
#include "a2name.h"
#include "extract.h"
#include "giop.h"


#ifndef timeradd
#define timeradd(a, b, result)   \
  do {      \
    (result)->tv_sec = (a)->tv_sec + (b)->tv_sec;                           \
    (result)->tv_usec = (a)->tv_usec + (b)->tv_usec;                        \
    if ((result)->tv_usec >= 1000000)                                       \
      {   \
      ++(result)->tv_sec;                                                   \
      (result)->tv_usec -= 1000000;                                         \
      }  \
  } while (0)

#endif

#ifndef timersub
#define timersub(a, b, result)  \
  do {  \
    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;                           \
    (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;                        \
    if ((result)->tv_usec < 0) {                                            \
      --(result)->tv_sec;                                                   \
      (result)->tv_usec += 1000000;                                         \
    }                                                                       \
  } while (0)
#endif


/*
 * The entries below are per 'stream', not per conversation
 * The src host, src port, dst host and dst port must all
 * match.
 */
struct iiop_tha {
       struct in_addr src;
       struct in_addr dst;
       u_short        src_port;
       u_short        dst_port;
     };

struct iiop_msg {
       struct iiop_msg  *nxt;
       struct timeval    first;
       struct timeval    last;
       struct iiop_tha   addr;
       struct MessageHeader_1_x *hdr;
     };

struct iiop_stream {
       struct iiop_stream *nxt;
       struct iiop_tha     addr;

       /* List of iiop messages for this stream */
       struct iiop_msg *msg;

       /* State of the current message for this stream */
       u_int size;
       u_int seq;
       u_int consumed;
     };

static struct iiop_msg    *requests;
static struct iiop_stream *iiop_streams = NULL;

static __inline u_int iiop_size (const struct MessageHeader_1_x *header)
{
  u_int32_t message_size = header->messageSize;

  return gtohl (message_size, header->flags);
}

static __inline u_int32_t giop_reply_id (const struct MessageHeader_1_x *header)
{
  const struct Sequence    *seq = (struct Sequence*) &header->payload;
  const struct ReplyHeader *repHeader;
  u_char *data = (u_char*) &seq->payload;
  int     i;

  for (i = 0; i < seq->numElements; i++)
     data = skip_context (data);

  repHeader = (struct ReplyHeader*) data;
  return (repHeader->request_id);
}

static __inline u_int32_t giop_request_id (const struct MessageHeader_1_x *header)
{
  const struct Sequence          *seq = (struct Sequence*) &header->payload;
  const struct RequestHeader_1_0 *reqHeader_1_0;
  const struct RequestHeader_1_1 *reqHeader_1_1;
  u_char *data = (u_char*) &seq->payload;
  int     version = header->GIOPVersion.major << 8 | header->GIOPVersion.minor;
  int     i;

  for (i = 0; i < seq->numElements; i++)
     data = skip_context (data);

  switch (version)
  {
    case 0x0100:            /* 1.0 */
         reqHeader_1_0 = (struct RequestHeader_1_0 *) (data);
         return (reqHeader_1_0->request_id);

    case 0x0101:            /* 1.1 */
         reqHeader_1_1 = (struct RequestHeader_1_1 *) (data);
         return (reqHeader_1_1->request_id);
  } 
  return (-1);
}

static __inline int is_oneway (const struct MessageHeader_1_x *header)
{
  const struct Sequence          *seq = (struct Sequence*) &header->payload;
  const struct RequestHeader_1_0 *reqHeader_1_0;
  const struct RequestHeader_1_1 *reqHeader_1_1;
  u_char *data = (u_char*) &seq->payload;
  int    version = header->GIOPVersion.major << 8 | header->GIOPVersion.minor;
  int    i;

  for (i = 0; i < seq->numElements; i++)
     data = skip_context (data);

  switch (version)
  {
    case 0x0100:            /* 1.0 */
         reqHeader_1_0 = (struct RequestHeader_1_0*) data;
         return (!reqHeader_1_0->response_expected);

    case 0x0101:            /* 1.1 */
         reqHeader_1_1 = (struct RequestHeader_1_1*) data;
         return (!reqHeader_1_1->response_expected);
  }
  return (0);
}

static void print_iiop_message (const struct MessageHeader_1_x *header)
{
  const struct Sequence          *seq = (struct Sequence*) &header->payload;
  const struct RequestHeader_1_0 *reqHeader_1_0;
  const struct RequestHeader_1_1 *reqHeader_1_1;
  const OctetSequence            *octetSeq = NULL;
  u_char *data = (u_char*) &seq->payload;
  int     version = header->GIOPVersion.major << 8 | header->GIOPVersion.minor;
  int     i;

  for (i = 0; i < seq->numElements; i++)
     data = skip_context (data);

  switch (version)
  {
    case 0x0100:            /* 1.0 */
         reqHeader_1_0 = (struct RequestHeader_1_0*) data;
         data = print_object_key (reqHeader_1_0->object_key._length,
                                  (u_char*) &reqHeader_1_0->object_key._buffer);
         octetSeq = (OctetSequence*) data; /* Should be String, but... */
         break;

    case 0x0101:            /* 1.1 */
         reqHeader_1_1 = (struct RequestHeader_1_1*) data;
         data = print_object_key (reqHeader_1_1->object_key._length,
                                  (u_char*) &reqHeader_1_1->object_key._buffer);
         octetSeq = (OctetSequence*) data; /* Should be String, but... */
         break;

    default:
         PRINTF ("version %d?", version);
         return;
  }
  PRINTF ("\"%s\" ", &octetSeq->_buffer); /* Already null terminated */

}

static void ts_diff_print (const struct timeval *tvp1, const struct timeval *tvp2)
{
  struct timeval diff;
  struct timeval *tvp = &diff;

  timersub (tvp1, tvp2, tvp);

  /* ts_print accomodates for timezone, so adjust here too
   */
  tvp->tv_sec -= thiszone;
  ts_print (tvp);
}


static void free_msg (struct iiop_msg *msg)
{
  free (msg->hdr);
  free (msg);
}


static void free_request (struct iiop_msg *request)
{
  if (request == requests)
     requests = request->nxt;
  else
  {
    struct iiop_msg *prev = requests;
    struct iiop_msg *req  = requests->nxt;

    while (req)
    {
      if (req == request)
      {
        prev->nxt = req->nxt;
        break;
      }
      prev = req;
      req = req->nxt;
    }
  }
  free_msg (request);
}

static void print_corba_call (struct iiop_msg *request,
                              struct iiop_msg *reply,
                              struct timeval  *reply_first,
                              struct timeval  *reply_last)
{
  struct MessageHeader_1_x *hdr = request->hdr;
  int    id     = giop_request_id (hdr);
  int    length = hdr->messageSize;

  /* Length needs to be incremented by the reply length
   */
  if (reply)
     length += reply->hdr->messageSize;

  PUTCHAR ('\n');
  if (reply)
       ts_print (reply_last);
  else ts_print (&request->last);

  PRINTF ("%s.%d > %s.%d: Corba %d.%d %d ",
          ipaddr_string (&request->addr.src), request->addr.src_port,
          ipaddr_string (&request->addr.dst), request->addr.dst_port,
          hdr->GIOPVersion.major, hdr->GIOPVersion.minor, length);

  if (reply)
       PRINTF ("Call %d ", id);
  else PRINTF ("Oneway %d ", id);

  print_iiop_message (request->hdr);

  /* Call-duration
   */
  if (reply)
       ts_diff_print (reply_last, &request->first);
  else ts_diff_print (&request->last, &request->first);

  if (reply)
  {
    /* request-length, reply-length, request-duration, reply-duration
     * Server-duration
      */
    PRINTF ("%d %d ", hdr->messageSize, reply->hdr->messageSize);
    ts_diff_print (&request->last, &request->first);
    ts_diff_print (reply_last, reply_first);
    ts_diff_print (reply_first, &request->last);
  }
  else
  {
    /* request-length, NA, request-duration, NA, NA
     */
    PRINTF ("%d NA ", hdr->messageSize);
    ts_diff_print (&request->last, &request->first);
    PUTS ("NA NA");
  }
  free_request (request);
}


static void print_unmatched_corba_request (struct iiop_msg *request)
{
  struct MessageHeader_1_x *hdr = request->hdr;
  int    id     = giop_request_id (hdr);
  int    length = hdr->messageSize;

  PUTCHAR ('\n');
  ts_print (&request->last);
  PRINTF ("%s.%d > %s.%d: Corba %d.%d ",
          ipaddr_string (&request->addr.src), request->addr.src_port,
          ipaddr_string (&request->addr.dst), request->addr.dst_port,
          hdr->GIOPVersion.major, hdr->GIOPVersion.minor);

  /* Total length is missing
   */
  PRINTF ("missing No_Reply %d ", id);
  print_iiop_message (request->hdr);

  /* Call duration is missing
   */
  PRINTF ("missing ");

  /* request-length, missing, request-duration, missing
   */
  PRINTF ("%d missing ", length);
  ts_diff_print (&request->last, &request->first);
  PUTS ("missing missing");
}

static void print_unmatched_corba_reply (struct iiop_msg *reply,
                                         u_int32_t id,
                                         struct iiop_tha *request_tha,
                                         struct timeval  *reply_first,
                                         struct timeval  *reply_last)
{
  struct MessageHeader_1_x *hdr = reply->hdr;

  PUTCHAR ('\n');
  ts_print (reply_last);
  PRINTF ("%s.%d > %s.%d: Corba %d.%d ",
          ipaddr_string (&request_tha->src), request_tha->src_port,
          ipaddr_string (&request_tha->dst), request_tha->dst_port,
          hdr->GIOPVersion.major, hdr->GIOPVersion.minor);

  /* Length is missing
   */
  PRINTF ("missing No_Request %d ", id);

  /* Three unavailable fields: object-key method call-duration
   */
  PUTS ("missing missing missing ");

  /* missing, reply-length, missing, reply-duration
   */
  PRINTF ("missing %d missing ", reply->hdr->messageSize);
  ts_diff_print (reply_last, reply_first);
  PUTS ("missing");
} 
  
/*
 * For now, only one prune per cycle
 */
static void prune_old_requests (struct timeval *time)
{
  struct iiop_msg    *request = requests;
  static const u_int  timeout = 30;   /* user-settable? */
  u_int  min = time->tv_sec - timeout;

  while (request)
  {
    if (request->last.tv_sec < min)
    {
      print_unmatched_corba_request (request);
      free_request (request);
      break;
    }
    request = request->nxt;
  }
}

static void reply_postprint (struct iiop_msg *reply,
                             struct iiop_tha *reply_tha,
                             u_int32_t id,
                             struct timeval *reply_first,
                             struct timeval *reply_last)
{
  struct iiop_tha  request_tha;
  struct iiop_msg *request = requests;

  request_tha.src = reply_tha->dst;
  request_tha.dst = reply_tha->src;
  request_tha.src_port = reply_tha->dst_port;
  request_tha.dst_port = reply_tha->src_port;
  while (request)
  {
    if (!memcmp (&request_tha, &request->addr, sizeof(request_tha)) &&
        id == giop_request_id(request->hdr))
      break;
    request = request->nxt;
  }

  if (request)
  {
    print_corba_call (request, reply, reply_first, reply_last);
  }
  else
  {
    /* No matching request */
    print_unmatched_corba_reply (reply, id, &request_tha,
                                 reply_first, reply_last);
  }
  prune_old_requests (reply_last);
}


/*
 * Two important points to note about print_corba:
 *
 * 1.  The byte vector passed in to this function is
 *     from tcpdump's main loop, and will get clobbered
 *     on the next cycle.  Since we need to save it, in
 *     order to frame iiop and match requests and replies,
 *     copies are made for Request and Reply types.
 *
 * 2,  The giop printer modifies certain numeric fields in
 *     place.  It's important not to use any of those fields
 *     until this in-place modification has been done.
 *
 * res, 13-April-1999
 */
void print_corba (const u_char *bp, u_int length, const u_char *bp2)
{
  const struct tcphdr *tp;
  const struct ip     *ip;
  struct iiop_tha      tha;
  struct iiop_stream  *iiop_th;

  u_short   sport, dport;
  u_int32_t seq;
  int       inMessage, payloadLength;
  u_char   *giop;

  tp = (struct tcphdr*) bp;
  ip = (struct ip*) bp2;

  sport = ntohs (tp->th_sport);
  dport = ntohs (tp->th_dport);
  seq   = ntohl (tp->th_seq);

  tha.src      = ip->ip_src;
  tha.dst      = ip->ip_dst;
  tha.src_port = sport;
  tha.dst_port = dport;

  for (iiop_th = iiop_streams; iiop_th; iiop_th = iiop_th->nxt)
  {
    if (!memcmp(&tha, &iiop_th->addr, sizeof(iiop_th->addr)))
      break;
  }

  if (!iiop_th)
  {
    iiop_th = (struct iiop_stream*) calloc (1, sizeof(*iiop_th));
    if (!iiop_th)
       error ("tcp_print: calloc");
    iiop_th->addr     = tha;
    iiop_th->consumed = iiop_th->size = 0;
    iiop_th->nxt      = iiop_streams;
    iiop_streams      = iiop_th;
  }

  inMessage = iiop_th->msg && (iiop_th->consumed < iiop_th->size);

  /* Only do the GIOP stuff if there's actually data
   * res 3-Nov-98
   *
   * Be sure to allow for non-zero options
   */
  payloadLength = length - 4*tp->th_off;

  if (inMessage)
  {
    int overflow;

    iiop_th->consumed = seq + payloadLength - iiop_th->seq;

#if 0
    PRINTF (" seq=%d consumed=%d size=%d ",
            iiop_th->seq, iiop_th->consumed, iiop_th->size);
#endif
    overflow = iiop_th->consumed - iiop_th->size;
    if (overflow >= payloadLength)
    {
      /* I'm assuming this means that data was dropped from the trace
       * and we're picking up somewhere after the end of the pending
       * message.
       */
      PUTS ("<iiop_incomplete> ");
      if (iiop_th->msg->hdr->messageType == Reply)
      {
        free_msg (iiop_th->msg);
      }
      else if (iiop_th->msg->hdr->messageType == Request)
      {
        if (is_oneway ((iiop_th->msg->hdr)))
          free_request (iiop_th->msg);
      }
      iiop_th->msg  = 0;
      iiop_th->size = iiop_th->consumed = 0;
      inMessage = 0;
    }
    else if (overflow >= 0)
    {
      /* If overflow is > 0, we could in theory look for GIOP at that
       * point in the current packet.
       */
      struct iiop_msg *msg = iiop_th->msg;

      msg->last.tv_sec  = PacketTimestamp->tv_sec;
      msg->last.tv_usec = PacketTimestamp->tv_usec;
      PUTS ("iiop ");
      corba_giop_print (msg->hdr);

      /* Duration of extended message
       */
      ts_diff_print (PacketTimestamp, &msg->first);
      if (msg->hdr->messageType == Reply)
      {
        u_int id = giop_reply_id (msg->hdr);

        reply_postprint (msg, &iiop_th->addr, id, &msg->first, &msg->last);
        free_msg (msg);
      }
      else if (msg->hdr->messageType == Request)
      {
        if (is_oneway (msg->hdr))
           print_corba_call (msg, 0, 0, 0);
      }
      iiop_th->msg  = 0;
      iiop_th->size = iiop_th->consumed = 0;
    }
  }

  if (payloadLength > 0 && !inMessage)
  {
    /* Be sure to allow for non-zero options
     */
    giop = (u_char*)tp + 4 * tp->th_off;

    if (*((u_int32_t*)giop) == GIOP_MAGIC)
    {
      iiop_th->msg = 0;

      if (snaplen > 80)
      {
        struct MessageHeader_1_x *header = (struct MessageHeader_1_x *) giop;
        struct iiop_msg          *msg    = NULL;
        u_int  msg_size                  = iiop_size (header);
        int    message_remaining         = msg_size - payloadLength + 12;

        iiop_th->size = msg_size;

        /* Don't count the initial 12 bytes of the iiop header
         */
        iiop_th->consumed = payloadLength;
        iiop_th->seq = seq + 12;

        if (header->messageType == Request)
        {
          int oneway = is_oneway (header);

          msg = calloc (1, sizeof(struct iiop_msg));

          msg->first.tv_sec  = PacketTimestamp->tv_sec;
          msg->first.tv_usec = PacketTimestamp->tv_usec;
          msg->last.tv_sec   = PacketTimestamp->tv_sec;
          msg->last.tv_usec  = PacketTimestamp->tv_usec;
          msg->hdr = calloc (1, payloadLength);
          memcpy (msg->hdr, header, payloadLength);
          msg->addr = iiop_th->addr;
          msg->nxt  = requests;
          requests  = msg;
          iiop_th->msg = msg;
          if (message_remaining > 0)
          {
            PUTS ("<request_begin>");
          }
          else
          {
            PUTS ("iiop ");
            corba_giop_print (msg->hdr);
            PUTS ("00:00:00.0 ");
            if (oneway)
            {
              print_corba_call (msg, 0, 0, 0);
              iiop_th->msg = 0;
            }
          }
        }
        else if (header->messageType == Reply)
        {
          if (message_remaining == 0)
          {
            u_int id;

            msg = calloc (1, sizeof(struct iiop_msg));

            msg->hdr = calloc (1, payloadLength);
            memcpy (msg->hdr, header, payloadLength);
            PUTS ("iiop ");
            corba_giop_print (msg->hdr);
            id = giop_reply_id (msg->hdr);
            PUTS ("00:00:00.0 ");
            reply_postprint (msg, &iiop_th->addr, id,
                             (struct timeval*)PacketTimestamp,
                             (struct timeval*)PacketTimestamp);
            free_msg (msg);
          }
          else
          {
            msg = calloc (1, sizeof(struct iiop_msg));

            msg->first.tv_sec  = PacketTimestamp->tv_sec;
            msg->first.tv_usec = PacketTimestamp->tv_usec;
            msg->hdr = calloc (1, payloadLength);
            memcpy (msg->hdr, header, payloadLength);
            iiop_th->msg = msg;
            PUTS ("<reply_begin>");
          }
        }
        else
        {
          PUTS ("iiop ");
          corba_giop_print (header);
          PUTS ("00:00:00.0 ");
        }
      }
      else
      {
        PUTS ("<iiop_truncated>");
      }
    }
  }
}
