/* 
 * Copyright (c) 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995, 1996
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <limits.h>

#include <sys/param.h>
#include <sys/socket.h>

#include "a2name.h"
#include "interfac.h"
#include "extract.h"            /* must come after interface.h */
#include "ip.h"


#ifndef IN_CLASSD
#define IN_CLASSD(i) (((int32_t)(i) & 0xf0000000) == 0xe0000000)
#endif


/* (following from ipmulti/mrouted/prune.h) */

/* 
 * The packet format for a traceroute request.
 */

struct tr_query {
       u_int32_t tr_src;            /* traceroute source */
       u_int32_t tr_dst;            /* traceroute destination */
       u_int32_t tr_raddr;          /* traceroute response address */
       u_int32_t tr_rttlqid;        /* response ttl and qid */
     };

#define TR_GETTTL(x)  (u_int32_t)(((x) >> 24) & 0xff)
#define TR_GETQID(x)  ((x) & 0x00ffffffUL)

/* 
 * Traceroute response format.  A traceroute response has a tr_query at the
 * beginning, followed by one tr_resp for each hop taken.
 */
struct tr_resp {
       u_int32_t tr_qarr;          /* query arrival time */
       u_int32_t tr_inaddr;        /* incoming interface address */
       u_int32_t tr_outaddr;       /* outgoing interface address */
       u_int32_t tr_rmtaddr;       /* parent address in source tree */
       u_int32_t tr_vifin;         /* input packet count on interface */
       u_int32_t tr_vifout;        /* output packet count on interface */
       u_int32_t tr_pktcnt;        /* total incoming packets for src-grp */
       u_int8_t  tr_rproto;        /* routing proto deployed on router */
       u_int8_t  tr_fttl;          /* ttl required to forward on outvif */
       u_int8_t  tr_smask;         /* subnet mask for src addr */
       u_int8_t  tr_rflags;        /* forwarding error codes */
     };

/* defs within mtrace
 */
#define TR_QUERY 1
#define TR_RESP 2

/* fields for tr_rflags (forwarding error codes)
 */
#define TR_NO_ERR       0
#define TR_WRONG_IF     1
#define TR_PRUNED       2
#define TR_OPRUNED      3
#define TR_SCOPED       4
#define TR_NO_RTE       5
#define TR_NO_FWD       7
#define TR_NO_SPACE     0x81
#define TR_OLD_ROUTER   0x82

/* fields for tr_rproto (routing protocol)
 */
#define TR_PROTO_DVMRP  1
#define TR_PROTO_MOSPF  2
#define TR_PROTO_PIM    3
#define TR_PROTO_CBT    4

static void print_mtrace (const u_char *bp, u_int len)
{
  struct tr_query *tr = (struct tr_query*) (bp + 8);

  PRINTF ("mtrace %lu: %s to %s reply-to %s",
          (u_long) TR_GETQID(ntohl(tr->tr_rttlqid)),
          ipaddr_string (&tr->tr_src),
          ipaddr_string (&tr->tr_dst),
          ipaddr_string (&tr->tr_raddr));

  if (IN_CLASSD (ntohl (tr->tr_raddr)))
     PRINTF (" with-ttl %d", TR_GETTTL (ntohl(tr->tr_rttlqid)));
  ARGSUSED (len);
}

static void print_mresp (const u_char *bp, u_int len)
{
  struct tr_query *tr = (struct tr_query *) (bp + 8);

  PRINTF ("mresp %lu: %s to %s reply-to %s",
          (u_long) TR_GETQID(ntohl(tr->tr_rttlqid)),
          ipaddr_string (&tr->tr_src),
          ipaddr_string (&tr->tr_dst),
          ipaddr_string (&tr->tr_raddr));

  if (IN_CLASSD (ntohl (tr->tr_raddr)))
     PRINTF (" with-ttl %d", TR_GETTTL (ntohl (tr->tr_rttlqid)));
  ARGSUSED (len);
}

static void igmp_print (const u_char *bp, u_int len, const u_char *bp2)
{
  const struct ip *ip = (const struct ip*) bp2;

  PRINTF ("%s > %s: ",
          ipaddr_string(&ip->ip_src),
          ipaddr_string(&ip->ip_dst));

  if (qflag)
  {
    PUTS ("igmp");
    return;
  }

  TCHECK2 (bp[0], 8);
  switch (bp[0])
  {
    case 0x11:
         PRINTF ("igmp %s query", bp[1] ? "v2" : "v1");
         if (bp[1] && bp[1] != 100)
            PRINTF (" [intvl %d]", bp[1]);
         PUTS ("igmp query");
         if (EXTRACT_32BITS (&bp[4]))
            PRINTF (" [gaddr %s]", ipaddr_string (&bp[4]));
         if (len != 8)
            PRINTF (" [len %d]", len);
         break;

    case 0x12:
    case 0x16:
         PRINTF ("igmp %s report %s", (bp[0] & 0x0f) == 6 ? "v2" : "v1", ipaddr_string (&bp[4]));
         if (len != 8)
            PRINTF (" [len %d]", len);
         if (bp[1])
            PRINTF (" [b1=0x%x]", bp[1]);
         break;

    case 0x17:
         PRINTF ("igmp leave %s", ipaddr_string (&bp[4]));
         if (len != 8)
            PRINTF (" [len %d]", len);
         if (bp[1])
            PRINTF (" [b1=0x%x]", bp[1]);
         break;

    case 0x13:
         PUTS ("igmp dvmrp");
         if (len < 8)
              PRINTF (" [len %d]", len);
         else dvmrp_print (bp, len);
         break;

    case 0x14:
         PUTS ("igmp pimv1");
         pimv1_print (bp, len);
         break;

    case 0x1e:
         print_mresp (bp, len);
         break;

    case 0x1f:
         print_mtrace (bp, len);
         break;

    default:
         PRINTF ("igmp-%d", bp[0] & 0xf);
         if (bp[1])
            PRINTF (" [b1=0x%02x]", bp[1]);
         break;
  }

  if (vflag && TTEST2 (bp[0], len))
  {
    /* Check the IGMP checksum */
    if (in_cksum ((u_short*)bp, len, 0))
       PRINTF (" bad igmp cksum %x!", EXTRACT_16BITS (&bp[2]));
  }
  return;

trunc:
  PUTS ("[|igmp]");
}


/* 
 * print the recorded route in an IP RR, LSRR or SSRR option.
 */
static void ip_printroute (const char *type, const u_char * cp, u_int length)
{
  u_int ptr = cp[2] - 1;
  u_int len;

  PRINTF (" %s{", type);
  if ((length + 1) & 3)
     PRINTF (" [bad length %d]", length);
  if (ptr < 3 || ((ptr + 1) & 3) || ptr > length + 1)
     PRINTF (" [bad ptr %d]", cp[2]);

  type = "";
  for (len = 3; len < length; len += 4)
  {
    if (ptr == len)
       type = "#";
    PRINTF ("%s%s", type, ipaddr_string (&cp[len]));
    type = " ";
  }
  PRINTF ("%s}", ptr == len ? "#" : "");
}

/*
 * print the timestamp in the IPOPT_TS option
 */
static void ip_printts (const u_char *cp, u_int length)
{
  u_int ptr = cp[2] - 1;
  u_int len = 0;
  int   hoplen;
  char *type;

  PUTS (" TS{");
  hoplen = ((cp[3] & 0xF) != IPOPT_TS_TSONLY) ? 8 : 4;

  if ((length - 4) & (hoplen-1))
     PRINTF ("[bad length %d]", length);

  if (ptr < 4 || ((ptr - 4) & (hoplen-1)) || ptr > length + 1)
     PRINTF ("[bad ptr %d]", cp[2]);

  switch (cp[3] & 0xF)
  {
    case IPOPT_TS_TSONLY:
         PUTS ("TSONLY");
         break;
    case IPOPT_TS_TSANDADDR:
         PUTS ("TS+ADDR");
         break;

        /*
         * prespecified should really be 3, but some ones might send 2
         * instead, and the IPOPT_TS_PRESPEC constant can apparently
         * have both values, so we have to hard-code it here.
         */
    case 2:
         PUTS ("PRESPEC2.0");
         break;
    case IPOPT_TS_PRESPEC:
         PUTS ("PRESPEC");
         break;
    default:
         PRINTF ("[bad ts type %d]", cp[3] & 0xF);
         goto done;
  }

  type = " ";
  for (len = 4; len < length; len += hoplen)
  {
    if (ptr == len)
       type = " ^ ";
    PRINTF ("%s%ld@%s", type, EXTRACT_32BITS (&cp[len+hoplen-4]),
            hoplen != 8 ? "" : ipaddr_string (&cp[len]));
    type = " ";
  }

done:
  PRINTF ("%s", ptr == len ? " ^ " : "");

  if (cp[3]>>4)
       PRINTF (" [%d hops not recorded]} ", cp[3] >> 4);
  else PUTCHAR ('}');
}


/* 
 * print IP options.
 */
static void ip_optprint (const u_char * cp, u_int length)
{
  u_int len;

  for (; length > 0; cp += len, length -= len)
  {
    int tt = *cp;

    if (tt == IPOPT_NOP || tt == IPOPT_EOL)
       len = 1;
    else
    {
      if (&cp[1] >= snapend)
      {
        PUTS ("[|ip]");
        return;
      }
      len = cp[1];
    }
    if (len <= 0)
    {
      PRINTF ("[|ip op len %d]", len);
      return;
    }
    if (&cp[1] >= snapend || cp + len > snapend)
    {
      PUTS ("[|ip]");
      return;
    }
    switch (tt)
    {
      case IPOPT_EOL:
           PUTS (" EOL");
           if (length > 1)
              PRINTF ("-%d", length - 1);
           return;

      case IPOPT_NOP:
           PUTS (" NOP");
           break;

      case IPOPT_TS:
           ip_printts(cp, len);
           break;

      case IPOPT_SECURITY:
           PRINTF (" SECURITY{%d}", len);
           break;

      case IPOPT_RR:
           ip_printroute ("RR", cp, len);
           break;

      case IPOPT_SSRR:
           ip_printroute ("SSRR", cp, len);
           break;

      case IPOPT_LSRR:
           ip_printroute ("LSRR", cp, len);
           break;

#ifndef IPOPT_RA
#define IPOPT_RA 148            /* router alert */
#endif
      case IPOPT_RA:
           PUTS (" RA");
           if (len != 4)
              PRINTF ("{%d}", len);
           else if (cp[2] || cp[3])
              PRINTF ("%d.%d", cp[2], cp[3]);
           break;

      default:
           PRINTF (" IPOPT-%d{%d}", cp[0], len);
           break;
    }
  }
}

#if !defined(USE_FAST_CKSUM) /* if not using fast version in libwatt.a */
/* 
 * compute an IP header checksum.
 * don't modifiy the packet.
 */
u_short in_cksum (register u_short *addr, register int len, u_short csum)
{
   register int     nleft = len;
   register u_short *w    = addr;
   register u_short answer;
   register int sum = csum;

  /*
   * Our algorithm is simple, using a 32 bit accumulator (sum),
   * we add sequential 16 bit words to it, and at the end, fold
   * back all the carry bits from the top 16 bits into the lower
   * 16 bits.
   */
  while (nleft > 1)
  {
    sum += *w++;
    nleft -= 2;
  }

  /* mop up an odd byte, if necessary
   */
  if (nleft == 1)
     sum += htons (*(u_char*)w << 8);

  /* add back carry outs from top 16 bits to low 16 bits
   */
  sum = (sum >> 16) + (sum & 0xffff); /* add hi 16 to low 16 */
  sum += (sum >> 16);                 /* add carry */
  answer = ~sum;                      /* truncate to 16 bits */
  return (answer);
}
#endif


/* 
 * print an IP datagram.
 */
void ip_print (const u_char *bp, u_int length)
{
  const  struct ip *ip = (const struct ip *) bp;
  const  u_char *cp;
  u_int  hlen, len, len0, off;
  u_char nh;
  u_long advance;
  int    enh;        /* encapsulated header */

#ifdef TCPDUMP_ALIGN
  /*
   * If the IP header is not aligned, copy into abuf.
   * This will never happen with BPF.  It does happen raw packet
   * dumps from -r.
   */
  if ((long) ip & 3)
  {
    static u_char *abuf = NULL;
    static int didwarn = 0;

    if (abuf == NULL)
    {
      abuf = malloc (snaplen);
      if (abuf == NULL)
        error ("ip_print: malloc");
    }
    memcpy (abuf, (char*)ip, min(length, snaplen));
    snapend += abuf - (u_char *) ip;
    packetp = abuf;
    ip = (struct ip *) abuf;
    /* We really want libpcap to give us aligned packets */
    if (!didwarn)
    {
      warning ("compensating for unaligned libpcap packets");
      ++didwarn;
    }
  }
#endif

  if ((u_char*)(ip+1) > snapend)
  {
    PUTS ("[|ip]");
    return;
  }

  if (length < sizeof(struct ip))
  {
    if (partial_frame)
         PRINTF ("[||ip]");
    else PRINTF ("truncated-ip %d\n", length);
    return;
  }

  if (IP_HL(ip) == 0)
  {
    PUTS ("bad packet - header length = 0");
    return;
  }

  hlen = IP_HL(ip) * 4;
  if (hlen < sizeof(struct ip))
  {
    PRINTF ("bad-hlen %d", hlen);
    return;
  }

  len = ntohs (ip->ip_len);
  if (length < len && !partial_frame)
     PRINTF ("truncated-ip - %d bytes missing!\n", len - length);

  len -= hlen;
  len0 = len;

  /*
   * If this is fragment zero, hand it to the next higher
   * level protocol.
   */
  off = ntohs (ip->ip_off);
  if (off & 0x2000)
     partial_frame++;

  if ((off & 0x1fff) == 0)
  {
    cp = (const u_char *) ip + hlen;
    nh = ip->ip_p;

    if (nh != IPPROTO_TCP && nh != IPPROTO_UDP)
       PRINTF ("%s > %s: ",
               ipaddr_string (&ip->ip_src),
               ipaddr_string (&ip->ip_dst));
again:
    switch (nh)
    {
      case IPPROTO_AH:
           nh = *cp;
           advance = ah_print (cp);
           cp  += advance;
           len -= advance;
           if (advance >= USHRT_MAX)
              break;
           goto again;

      case IPPROTO_ESP:
           advance = esp_print (cp, (const u_char*)ip, &enh);
           cp  += advance;
           len -= advance;
           if (enh < 0 || advance >= USHRT_MAX)
              break;
           nh = enh & 0xff;
           goto again;

      case IPPROTO_IPCOMP:
           advance = ipcomp_print (cp, (const u_char *) ip, &enh);
           cp  += advance;
           len -= advance;
           if (enh < 0 || advance >= USHRT_MAX)
              break;
           nh = enh & 0xff;
           goto again;

      case IPPROTO_TCP:
           tcp_print (cp, len, (const u_char *)ip, (off & ~0x6000));
           break;

      case IPPROTO_UDP:
           udp_print (cp, len, (const u_char *)ip, (off & ~0x6000));
           break;

      case IPPROTO_XTP:
           xtp_print (cp, len, (const u_char *)ip);
           break;

      case IPPROTO_ICMP:
           icmp_print (cp, len, (const u_char *)ip);
           break;

      case IPPROTO_IGRP:
           igrp_print (cp, len, (const u_char *) ip);
           break;

      case IPPROTO_ND:
           PRINTF (" nd %d", len);
           break;

      case IPPROTO_EGP:
           egp_print (cp, len, (const u_char *) ip);
           break;

      case IPPROTO_OSPF:
           ospf_print (cp, len, (const u_char *) ip);
           break;

      case IPPROTO_IGMP:
           igmp_print (cp, len, (const u_char *) ip);
           break;

      case IPPROTO_DVMRP:  /* multicast tunnel (ip-in-ip encapsulation) */
           ip_print (cp, len);
           if (!vflag)
           {
             PUTS (" (ipip)");
             return;
           }
           break;

#ifdef USE_INET6
      case IP6PROTO_ENCAP:     /* ip6-in-ip encapsulation */
      case IPPROTO_IPV6:
           ip6_print (cp, len);
           if (!vflag)
           {
             PUTS (" (encap)");
             return;
           }
           break;
#endif
      case IPPROTO_GRE:
           gre_print (cp, len);
           if (!vflag)
           {
             PUTS (" (gre encap)");
             return;
           }
           break;
 
      case IPPROTO_MOBILE:
           mobile_print (cp, len);
           if (!vflag)
           {
             PUTS (" (mobile encap)");
             return;
           }
           break;

      case IPPROTO_PIM:
           pim_print (cp, len);
           break;

      case IPPROTO_RSVP:
           PUTS (" rsvp");
#ifdef USE_RSVP
           ntoh_rsvp_packet (cp, len);
           rsvp_print_pkt (cp, len);
#endif
           break;

      case IPPROTO_AXIP:
           axip_print (cp, len, (const u_char *)ip);
           break;

#ifndef IPPROTO_VRRP
#define IPPROTO_VRRP	112
#endif
      case IPPROTO_VRRP:
           if (vflag)
              PRINTF ("vrrp %s > %s: ",
                      ipaddr_string (&ip->ip_src),
                      ipaddr_string (&ip->ip_dst));
           vrrp_print (cp, len, ip->ip_ttl);
           break;

      default:
           PRINTF (" ip-proto-%d %d", nh, len);
           break;
    }
  }
  if (off & 0x2000)
      partial_frame--;

  /* Ultra quiet now means that all this stuff should be suppressed
   */
  if (qflag > 1)
     return;

  /*
   * for fragmented datagrams, print id:size@offset.  On all
   * but the last stick a "+".  For unfragmented datagrams, note
   * the don't fragment flag.
   */
  len = len0;             /* get the original length */
  if (off & 0x3fff)
  {
    /*
     * if this isn't the first frag, we're missing the
     * next level protocol header.  print the ip addr.
     */
    if (off & 0x1fff)
       PRINTF ("%s > %s:", ipaddr_string (&ip->ip_src),
               ipaddr_string (&ip->ip_dst));
    PRINTF (" (frag %d:%d@%d%s)", ntohs (ip->ip_id), len,
            (off & 0x1fff) * 8,
            (off & IP_MF) ? "+" : "");
  }
  else if (off & IP_DF)
          PUTS (" (DF)");

  if (ip->ip_tos)
  {
    PRINTF (" [tos 0x%x", (int) ip->ip_tos);
    /* ECN bits */
    if (ip->ip_tos &0x01)
       PUTS (",CE");
    if (ip->ip_tos &0x02)
       PUTS (",ECT");
    PUTS ("] ");
  }

  if (ip->ip_ttl <= 1)
     PRINTF (" [ttl %d]", (int) ip->ip_ttl);

  if (vflag)
  {
    int  sum;
    char *sep = "";

    PUTS (" (");
    if (ip->ip_ttl > 1)
    {
      PRINTF ("%sttl %d", sep, (int)ip->ip_ttl);
      sep = ", ";
    }
    if ((off & 0x3fff) == 0)
    {
      PRINTF ("%sid %d", sep, (int)ntohs(ip->ip_id));
      sep = ", ";
    }
    if ((u_char*)ip + hlen <= snapend)
    {
      sum = in_cksum ((u_short*)ip, hlen, 0);
      if (sum != 0)
      {
        PRINTF ("%sbad cksum %x!", sep, ntohs(ip->ip_sum));
        sep = ", ";
      }
    }
    hlen -= sizeof(struct ip);
    if (hlen > 0)
    {
      PRINTF ("%soptlen=%d", sep, hlen);
      ip_optprint ((u_char *) (ip + 1), hlen);
    }
    PUTCHAR (')');
  }
}

void ipN_print (const u_char * bp, u_int length)
{
  struct ip *ip, hdr;

  ip = (struct ip *) bp;
  if (length < 4)
  {
    PRINTF ("truncated-ip %d", length);
    return;
  }
  memcpy (&hdr, (char*)ip, 4);
  switch (IP_V (&hdr))
  {
    case 4:
         ip_print (bp, length);
         return;
#ifdef USE_INET6
    case 6:
         ip6_print (bp, length);
         return;
#endif
    default:
         PRINTF ("unknown ip %d", IP_V(&hdr));
         return;
  }
}
