/*
 * Copyright (c) 1992, 1993, 1994, 1995, 1996
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * OSPF support contributed by Jeffrey Honig (jch@mitchell.cit.cornell.edu)
 */

#include <stdio.h>
#include <ctype.h>
#include <time.h>

#include "interfac.h"
#include "a2name.h"
#include "ip.h"
#include "ospf.h"

struct bits {
       u_int bit;
       const char *str;
     };

static const struct bits ospf_option_bits[] = {
                    { OSPF_OPTION_T,  "T"  },
                    { OSPF_OPTION_E,  "E"  },
                    { OSPF_OPTION_MC, "MC" },
                    { 0,              NULL }
                  };

static const struct bits ospf_rla_flag_bits[] = {
                    { RLA_FLAG_B,  "B"  },
                    { RLA_FLAG_E,  "E"  },
                    { RLA_FLAG_W1, "W1" },
                    { RLA_FLAG_W2, "W2" },
                    { 0,           NULL }
                  };

static const char *ospf_types[OSPF_TYPE_MAX] = {
                   NULL, "hello", "dd",
                   "ls_req", "ls_upd", "ls_ack"
                 };

static void ospf_print_seqage (u_int seq, time_t us)
{
  time_t sec  = us % 60;
  time_t mins = (us / 60) % 60;
  time_t hour = us / 3600;

  PRINTF (" S %X age ", seq);
  if (hour)
       PRINTF ("%u:%02u:%02u", (u_int)hour, (u_int)mins, (u_int)sec);
  else if (mins)
       PRINTF ("%u:%02u", (u_int)mins, (u_int)sec);
  else PRINTF ("%u", (u_int)sec);
}


static void ospf_print_bits (const struct bits *bp, u_char options)
{
  char sep = ' ';

  do
  {
    if (options & bp->bit)
    {
      PRINTF ("%c%s", sep, bp->str);
      sep = '/';
    }
  }
  while ((++bp)->bit);
}


#define LS_PRINT(lsp, type) \
  switch (type) {           \
    case LS_TYPE_ROUTER:    \
         PRINTF (" rtr %s ", ipaddr_string(&lsp->ls_router)); \
         break;             \
    case LS_TYPE_NETWORK:   \
         PRINTF (" net dr %s if %s", ipaddr_string(&lsp->ls_router), ipaddr_string(&lsp->ls_stateid)); \
         break;             \
    case LS_TYPE_SUM_IP:    \
         PRINTF (" sum %s abr %s", ipaddr_string(&lsp->ls_stateid), ipaddr_string(&lsp->ls_router)); \
         break;             \
    case LS_TYPE_SUM_ABR:   \
         PRINTF (" abr %s rtr %s", ipaddr_string(&lsp->ls_router), ipaddr_string(&lsp->ls_stateid)); \
         break;             \
    case LS_TYPE_ASE:       \
         PRINTF (" ase %s asbr %s", ipaddr_string(&lsp->ls_stateid), ipaddr_string(&lsp->ls_router)); \
         break;             \
    case LS_TYPE_GROUP:     \
         PRINTF (" group %s rtr %s", ipaddr_string(&lsp->ls_stateid), ipaddr_string(&lsp->ls_router)); \
         break;             \
  }

static int ospf_print_lshdr (const struct lsa_hdr *lshp, const caddr_t end)
{
  if ((caddr_t) (lshp + 1) > end)
     return (1);

  PUTS (" {");

  if (!lshp->ls_type || lshp->ls_type >= LS_TYPE_MAX)
  {
    PRINTF (" ??LS type %d?? }", lshp->ls_type);
    return (1);
  }

  ospf_print_bits (ospf_option_bits, lshp->ls_options);
  ospf_print_seqage (ntohl (lshp->ls_seq), ntohs (lshp->ls_age));

  LS_PRINT (lshp, lshp->ls_type);
  return (0);
}


/* 
 * Print a single link state advertisement.  If truncated return 1, else 0.
 */

static int ospf_print_lsa (const struct lsa *lsap, const caddr_t end)
{
  const char              *ls_end;
  const struct rlalink    *rlp;
  const struct tos_metric *tosp;
  const struct in_addr    *ap;
  const struct aslametric *almp;
  const struct mcla       *mcp;
  const u_int32_t         *lp;
  int   j, k;

  if (ospf_print_lshdr (&lsap->ls_hdr, end))
     return (1);

  ls_end = (const char *) lsap + ntohs (lsap->ls_hdr.ls_length);
  if (ls_end > (const char *) end)
  {
    PUTS (" }");
    return (1);
  }

  switch (lsap->ls_hdr.ls_type)
  {
    case LS_TYPE_ROUTER:
         ospf_print_bits (ospf_rla_flag_bits, lsap->lsa_un.un_rla.rla_flags);

         j = ntohs (lsap->lsa_un.un_rla.rla_count);
         rlp = lsap->lsa_un.un_rla.rla_link;
         while (j--)
         {
           struct rlalink *rln = (struct rlalink*) ((caddr_t)(rlp + 1) +
                                 ((rlp->link_toscount) * sizeof(struct tos_metric)));

           if ((const char *) rln > ls_end)
              break;
           PUTS (" {");

           switch (rlp->link_type)
           {
             case RLA_TYPE_VIRTUAL:
                  PUTS (" virt");
                  /* Fall through */

             case RLA_TYPE_ROUTER:
                  PRINTF (" nbrid %s if %s",
                          ipaddr_string (&rlp->link_id),
                          ipaddr_string (&rlp->link_data));
                  break;

             case RLA_TYPE_TRANSIT:
                  PRINTF (" dr %s if %s",
                          ipaddr_string (&rlp->link_id),
                          ipaddr_string (&rlp->link_data));
                  break;

             case RLA_TYPE_STUB:
                  PRINTF (" net %s mask %s",
                          ipaddr_string (&rlp->link_id),
                          ipaddr_string (&rlp->link_data));
                  break;

             default:
                  PRINTF (" ??RouterLinksType %d?? }", rlp->link_type);
                  return 0;
           }
           PRINTF (" tos 0 metric %d", ntohs (rlp->link_tos0metric));
           tosp = (struct tos_metric *) (sizeof(rlp->link_tos0metric) + (caddr_t)rlp);
           for (k = 0; k < rlp->link_toscount; k++, tosp++)
           {
             PRINTF (" tos %d metric %d",
                     tosp->tos_type, ntohs (tosp->tos_metric));
           }
           PUTS (" }");
           rlp = rln;
         }
         break;

    case LS_TYPE_NETWORK:
         PRINTF (" mask %s rtrs",
                 ipaddr_string (&lsap->lsa_un.un_nla.nla_mask));
         for (ap = lsap->lsa_un.un_nla.nla_router;
              (const char *) (ap + 1) <= ls_end;
              ap++)
         {
           PUTCHAR (' ');
           PUTS (ipaddr_string (ap));
         }
         break;

    case LS_TYPE_SUM_IP:
         PRINTF (" mask %s", ipaddr_string (&lsap->lsa_un.un_sla.sla_mask));
         /* Fall through */

    case LS_TYPE_SUM_ABR:
         for (lp = lsap->lsa_un.un_sla.sla_tosmetric;
              (const char *) (lp + 1) <= ls_end;
              lp++)
         {
           u_int ul = ntohl (*lp);
           PRINTF (" tos %d metric %d",
                   (ul & SLA_MASK_TOS) >> SLA_SHIFT_TOS,
                   ul & SLA_MASK_METRIC);
         }
         break;

    case LS_TYPE_ASE:
         PRINTF (" mask %s",
                 ipaddr_string (&lsap->lsa_un.un_asla.asla_mask));

         for (almp = lsap->lsa_un.un_asla.asla_metric;
              (const char *) (almp + 1) <= ls_end;
              almp++)
         {
           u_int ul = ntohl (almp->asla_tosmetric);

           PRINTF (" type %d tos %d metric %d",
                   (ul & ASLA_FLAG_EXTERNAL) ? 2 : 1,
                   (ul & ASLA_MASK_TOS) >> ASLA_SHIFT_TOS,
                   (ul & ASLA_MASK_METRIC));
           if (almp->asla_forward.s_addr)
              PRINTF (" forward %s", ipaddr_string (&almp->asla_forward));

           if (almp->asla_tag.s_addr)
              PRINTF (" tag %s", ipaddr_string (&almp->asla_tag));
         }
         break;

    case LS_TYPE_GROUP:
         /* Multicast extensions as of 23 July 1991 */
         for (mcp = lsap->lsa_un.un_mcla;
              (const char *) (mcp + 1) <= ls_end;
              mcp++)
         {
           switch (ntohl (mcp->mcla_vtype))
           {
             case MCLA_VERTEX_ROUTER:
                  PRINTF (" rtr rtrid %s",
                          ipaddr_string (&mcp->mcla_vid));
                  break;

             case MCLA_VERTEX_NETWORK:
                  PRINTF (" net dr %s",
                          ipaddr_string (&mcp->mcla_vid));
                  break;

             default:
                  PRINTF (" ??VertexType %u??",
                          (u_int) ntohl (mcp->mcla_vtype));
                  break;
           }
         }
  }

  PUTS (" }");
  return (0);
}


void ospf_print (const u_char *bp, u_int length, const u_char *bp2)
{
  const struct ospfhdr *op;
  const struct ip      *ip;
  const struct lsa     *lsap;
  const struct lsa_hdr *lshp;
  const struct in_addr *ap;
  const struct lsr     *lsrp;
  const caddr_t         end = (caddr_t) snapend;
  char  sep;
  int   i, j;

  op = (struct ospfhdr *) bp;
  ip = (struct ip *) bp2;
  /* Print the source and destination address */
  PRINTF ("%s > %s:",
          ipaddr_string (&ip->ip_src),
          ipaddr_string (&ip->ip_dst));

  if ((caddr_t) (&op->ospf_len + 1) > end)
     goto trunc_test;


  /* If the type is valid translate it, or just print the type */
  /* value.  If it's not valid, say so and return */
  if (op->ospf_type || op->ospf_type < OSPF_TYPE_MAX)
  {
    PRINTF (" OSPFv%d-%s %d:",
            op->ospf_version, ospf_types[op->ospf_type], length);
  }
  else
  {
    PRINTF (" ospf-v%d-??type %d?? %d:",
            op->ospf_version, op->ospf_type, length);
    return;
  }

  if (length != ntohs (op->ospf_len))
  {
    PRINTF (" ??len %d??", ntohs (op->ospf_len));
    goto trunc_test;
  }

  if ((caddr_t) (&op->ospf_routerid + 1) > end)
     goto trunc_test;

  /* Print the routerid if it is not the same as the source */
  if (ip->ip_src.s_addr != op->ospf_routerid.s_addr)
     PRINTF (" rtrid %s", ipaddr_string (&op->ospf_routerid));

  if ((caddr_t) (&op->ospf_areaid + 1) > end)
     goto trunc_test;

  if (op->ospf_areaid.s_addr)
       PRINTF (" area %s", ipaddr_string (&op->ospf_areaid));
  else PUTS   (" backbone");

  if ((caddr_t) (op->ospf_authdata + OSPF_AUTH_SIZE) > end)
     goto trunc_test;

  if (vflag)
  {
    /* Print authentication data (should we really do this?) */
    switch (ntohs (op->ospf_authtype))
    {
      case OSPF_AUTH_NONE:
           break;

      case OSPF_AUTH_SIMPLE:
           PUTS (" auth ");
           j = 0;
           for (i = 0; i < sizeof(op->ospf_authdata); i++)
           {
             if (!isprint (op->ospf_authdata[i]))
             {
               j = 1;
               break;
             }
           }
           if (j)
           {
             /* Print the auth-data as a string of octets */
             PRINTF ("%s.%s",
                     ipaddr_string ((struct in_addr *) op->ospf_authdata),
                     ipaddr_string ((struct in_addr *) &op->ospf_authdata[sizeof(struct in_addr)]));
           }
           else
           {
             /* Print the auth-data as a text string */
             PRINTF ("'%.8s'", op->ospf_authdata);
           }
           break;

      default:
           PRINTF (" ??authtype-%d??", ntohs (op->ospf_authtype));
           return;
    }
  }


  /* Do rest according to version.    */
  switch (op->ospf_version)
  {
    case 2:
         /* ospf version 2 */
         switch (op->ospf_type)
         {
           case OSPF_TYPE_UMD:  /* Rob Coltun's special monitoring packets; do nothing  */
                break;

           case OSPF_TYPE_HELLO:
                if ((caddr_t) (&op->ospf_hello.hello_deadint + 1) > end)
                   break;

                if (vflag)
                {
                  ospf_print_bits (ospf_option_bits,
                                   op->ospf_hello.hello_options);
                  PRINTF (" mask %s int %d pri %d dead %u",
                          ipaddr_string (&op->ospf_hello.hello_mask),
                          ntohs (op->ospf_hello.hello_helloint),
                          op->ospf_hello.hello_priority,
                          (u_int) ntohl (op->ospf_hello.hello_deadint));
                }

                if ((caddr_t)(&op->ospf_hello.hello_dr + 1) > end)
                   break;

                if (op->ospf_hello.hello_dr.s_addr)
                   PRINTF (" dr %s", ipaddr_string (&op->ospf_hello.hello_dr));

                if ((caddr_t) (&op->ospf_hello.hello_bdr + 1) > end)
                   break;

                if (op->ospf_hello.hello_bdr.s_addr)
                   PRINTF (" bdr %s", ipaddr_string (&op->ospf_hello.hello_bdr));

                if (vflag)
                {
                  if ((caddr_t) (op->ospf_hello.hello_neighbor + 1) > end)
                     break;

                  PUTS (" nbrs");
                  for (ap = op->ospf_hello.hello_neighbor;
                       (caddr_t) (ap + 1) <= end; ap++)
                  {
                    PUTCHAR (' ');
                    PUTS (ipaddr_string (ap));
                  }
                }
                break;          /* HELLO */

           case OSPF_TYPE_DB:
                if ((caddr_t) (&op->ospf_db.db_seq + 1) > end)
                   break;

                ospf_print_bits (ospf_option_bits, op->ospf_db.db_options);
                sep = ' ';
                if (op->ospf_db.db_flags & OSPF_DB_INIT)
                {
                  PRINTF ("%cI", sep);
                  sep = '/';
                }
                if (op->ospf_db.db_flags & OSPF_DB_MORE)
                {
                  PRINTF ("%cM", sep);
                  sep = '/';
                }
                if (op->ospf_db.db_flags & OSPF_DB_MASTER)
                {
                  PRINTF ("%cMS", sep);
                  sep = '/';
                }
                PRINTF (" S %X", (u_int)ntohl (op->ospf_db.db_seq));

                if (vflag)
                {
                  /* Print all the LS adv's */
                  lshp = op->ospf_db.db_lshdr;

                  while (!ospf_print_lshdr (lshp, end))
                  {
                    PUTS (" }");
                    lshp++;
                  }
                }
                break;

           case OSPF_TYPE_LSR:
                if (vflag)
                {
                  for (lsrp = op->ospf_lsr; (caddr_t)(lsrp + 1) <= end; lsrp++)
                  {
                    long type;

                    if ((caddr_t) (lsrp + 1) > end)
                       break;

                    PUTS (" {");
                    if ((type = ntohl (lsrp->ls_type)) != 0 || type >= LS_TYPE_MAX)
                    {
                      PRINTF (" ??LinkStateType %ld } }", type);
                      break;
                    }
                    LS_PRINT (lsrp, type);
                    PUTS (" }");
                  }
                }
                break;

           case OSPF_TYPE_LSU:
                if (vflag)
                {
                  lsap = op->ospf_lsu.lsu_lsa;
                  i = ntohl (op->ospf_lsu.lsu_count);

                  while (i-- &&
                         !ospf_print_lsa (lsap, end))
                     lsap = (struct lsa*) ((caddr_t)lsap +
                                           ntohs(lsap->ls_hdr.ls_length));
                }
                break;

           case OSPF_TYPE_LSA:
                if (vflag)
                {
                  lshp = op->ospf_lsa.lsa_lshdr;

                  while (!ospf_print_lshdr (lshp, end))
                  {
                    PUTS (" }");
                    lshp++;
                  }
                  break;
                }
         }
         break;

    default:
         PRINTF (" ospf [version %d]", op->ospf_version);
         break;
  }

trunc_test:
  if (snapend - bp < length)
     PUTS (" [|]");
}
