/*
 * Copyright (c) 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995, 1996
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* 
 * tcpdump - monitor tcp/ip traffic on an ethernet.
 *
 * First written in 1987 by Van Jacobson, Lawrence Berkeley Laboratory.
 * Mercilessly hacked and occasionally improved since then via the
 * combined efforts of Van, Steve McCanne and Craig Leres of LBL.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <setjmp.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pcap.h>
#include <tcp.h>

#ifdef __DJGPP__
#include <sys/exceptn.h>
#include <crt0.h>
#endif

#include "interfac.h"
#include "a2name.h"
#include "ethertyp.h"

int  aflag  = 0;                 /* translate network & broadcast addresses */
int  fflag  = 0;                 /* don't translate "foreign" IP address */
int  Lflag  = 0;                 /* list available data link types and exit */
int  Hflag  = 0;                 /* append resolved host names to hosts file */
int  nflag  = 0;                 /* leave addresses as numbers */
int  Nflag  = 0;                 /* remove domains from printed host names */
int  pflag  = 0;                 /* don't go promiscuous */
int  Pflag  = 0;                 /* dump protocol info */
int  qflag  = 0;                 /* quick (shorter) output */
int  tflag  = 1;                 /* print packet arrival time */
int  uflag  = 0;                 /* Print undecoded NFS handles */
int  eflag  = 0;                 /* print ethernet header */
int  vflag  = 0;                 /* verbose */
int  xflag  = 0;                 /* print packet in hex */
int  wflag  = 0;                 /* write packets to file */
int  Aflag  = 0;                 /* print packet in ascii */
int  xaflag = 0;                 /* print packet in hex & ascii */
int  Oflag  = 1;                 /* run filter code optimizer */
int  Sflag  = 0;                 /* print raw TCP sequence numbers */
int  sflag  = 0;                 /* use the libsmi to translate OIDs */
int  dflag  = 0;                 /* print filter code */
int  Dflag  = 0;                 /* dump device info */
int  Rflag  = 0;                 /* print packets read from file at real pace */

int  packettype;
int  messagetype;
int  break_mode = 1;             /* ^Break/^C mode (djgpp only) */

int  snaplen = DEFAULT_SNAPLEN;  /* Length of saved portion of packet. */
long thiszone;                   /* seconds offset from gmt to local time */

int  partial_frame;
WORD extracted_ethertype;

char program_name[_MAX_PATH];
char program_path[_MAX_PATH];
char *pcap_device = NULL;
char *espsecret = NULL;          /* ESP secret key */

DWORD localnet;
DWORD netmask;
WORD  protocol_b;

const struct timeval *PacketTimestamp;
const u_char *packetp;
const u_char *snapend;

static pcap_t *pd        = NULL;
static char   *RFileName = NULL;
static char   *WFileName = NULL;

static int dlt = -1;        /* if != -1, ask libpcap for the DLT it names */
static const char *dlt_name = NULL;

static struct bpf_program fcode;

static void program_usage (int details);
static void program_end   (int free_mem);
static void init_handlers (void);
static void dump_stats    (void);
static void dump_ascii    (const u_char *bp, u_int length);
static void dump_hex_ascii(const u_char *bp, u_int length);

struct printer {
       pcap_handler f;
       int          type;
     };

#ifndef DLT_ATM_RFC1483
#define DLT_ATM_RFC1483 11
#endif


static struct printer printers[] = {
              { ether_if_print, DLT_EN10MB      },
              { token_if_print, DLT_IEEE802     },
           // { lane_if_print,  DLT_LANE8023    },
           // { cip_if_print,   DLT_CIP         },
           // { cip_if_print,   DLT_ATM_CLIP    },
              { sl_if_print,    DLT_SLIP        },
              { sl_bs_if_print, DLT_SLIP_BSDOS  },
              { ppp_if_print,   DLT_PPP         },
              { ppp_bs_if_print,DLT_PPP_BSDOS   },
              { fddi_if_print,  DLT_FDDI        },
              { null_if_print,  DLT_NULL        },
           // { null_if_print,  DLT_LOOP        },
              { raw_if_print,   DLT_RAW         },
              { atm_if_print,   DLT_ATM_RFC1483 },
           // { chdlc_if_print, DLT_C_HDLC      },
              { ax25_if_print,  DLT_AX25        },
           // { sll_if_print,   DLT_LINUX_SLL   },
              { NULL,           0               },
            };

static pcap_handler lookup_printer (int type)
{
  struct printer *p;

  for (p = printers; p->f; ++p)
      if (type == p->type)
         return (p->f);

  PERROR (("unknown data link type 0x%x", type));
  return (NULL);
}

static unsigned short eth_proto_parse (char *id)
{
   if (!strcmp(id, "ip") || !strcmp(id, "ipv4"))
      return htons (ETHERTYPE_IP);
   if (!strcmp(id, "ipv6"))
      return htons (ETHERTYPE_IPV6);
   if (!strcmp(id, "arp"))
      return htons (ETHERTYPE_ARP);
   if (!strcmp(id, "rarp"))
      return htons (ETHERTYPE_RARP);
   if (!strcmp(id, "802.2"))
      return htons (ETHERTYPE_802_2);
   if (!strcmp(id, "802.3"))
      return htons (ETHERTYPE_802_3);
   if (!strcmp(id, "lat"))
      return htons (ETHERTYPE_LAT);
   if (!strcmp(id, "dec"))
      return htons (ETHERTYPE_DEC);
   if (!strcmp(id, "atalk"))
      return htons (ETHERTYPE_ATALK);
   if (!strcmp(id, "aarp"))
      return htons (ETHERTYPE_AARP);
   if (!strcmp(id, "ipx"))
      return htons (ETHERTYPE_IPX);
   if (!strcmp(id, "x25"))
      return htons (ETHERTYPE_X25);
   return htons (atoi(id));
}

static void show_dlts_and_exit (pcap_t *pd)
{
  int *dlts   = NULL;
  int  n_dlts = pcap_list_datalinks (pd, &dlts);

  if (n_dlts < 0)
     error ("%s", pcap_geterr(pd));

  if (n_dlts == 0 || !dlts)
     error ("No data link types.");

  fprintf (stderr, "Data link types (use option -y to set):\n");

  while (--n_dlts >= 0)
  {
    const char *name = pcap_datalink_val_to_name (dlts[n_dlts]);
    if (name)
    {
      fprintf (stderr, "  %s", name);

      /* OK, does tcpdump handle that type?
       */
      if (lookup_printer(dlts[n_dlts]) == NULL)
         fprintf (stderr, " (not supported)");
      putchar ('\n');
    }
    else
      fprintf (stderr, "  DLT %d (not supported)\n", dlts[n_dlts]);
  }
  free (dlts);
  exit (0);
}


/************************************************************************/

int main (int argc, char **argv)
{
  static char *infile = NULL;
  static int   count  = -1;
  static int   rc     = 0;

  pcap_handler printer;
  int          op;
  u_char      *pcap_userdata;
  char         ebuf[PCAP_ERRBUF_SIZE];
  char        *cmdbuf;
  char        *cp = strrchr (argv[0], SLASH);

  if (cp)
  {
    strcpy (program_name, cp+1);
    memcpy (&program_path, argv[0], cp-argv[0]+1);
    program_path [(int)(cp+2-argv[0])] = '\0';
  }
  else
  {
    strcpy (program_name, argv[0]);
    program_path[0] = '\0';
  }
  strlwr (program_name);
  strlwr (program_path);

  opterr = 1;

  while ((op = getopt(argc,argv,"aAb:Cc:dDeLE:fF:Hi:lm:M:nNOpPqRr:s:StT:uvw:xXyh?")) != EOF)
  {
    switch (op)
    {
      case 'a':
           ++aflag;
           break;

      case 'b':
           protocol_b = eth_proto_parse (optarg);
           break;

      case 'C':
           use_conio = 1;
           if (wflag)
              PERROR (("`-C' cannot be used with `-w' flag"));
           break;

      case 'c':
           count = atoi (optarg);
           if (count <= 0)
              PERROR (("invalid packet count %s", optarg));
           break;

      case 'd':
           ++dflag;
           break;

      case 'D':
           ++Dflag;
           break;

      case 'e':
           ++eflag;
           break;

      case 'L':
           Lflag++;
           break;

      case 'E':
#ifndef USE_SSL
           warning ("crypto code not compiled in");
#endif
           espsecret = optarg;
           break;

      case 'f':
           ++fflag;
           break;

      case 'F':
           infile = optarg;
           break;

      case 'H':
           ++Hflag;
           break;

      case 'i':
           pcap_device = optarg;
           break;

      case 'l':
           setvbuf (stdout, NULL, _IOLBF, 0);
           break;

      case 'm':
#ifdef LIBSMI
           if (!smiLoadModule(optarg))
              error ("could not load MIB module %s", optarg);
           sflag = 1;
#else
           fprintf (stderr, "%s: ignoring option `-m %s' ", program_name, optarg);
           fprintf (stderr, "(no libsmi support)\n");
#endif
           break;

      case 'M':
           if (!stricmp(optarg, "oncrpc"))
                messagetype = MT_ONCRPC;
           else if (!stricmp(optarg, "giop"))
                messagetype = MT_GIOP;
           else if (!stricmp(optarg, "w3ng"))
                messagetype = MT_W3NG;
           else if (!stricmp(optarg, "http"))
                messagetype = MT_HTTP;
           else PERROR (("Legal message types are: oncrpc, giop, w3ng and http"));
           break;

      case 'n':
           ++nflag;
           break;

      case 'N':
           ++Nflag;
           break;

      case 'O':
           Oflag = 0;
           break;

      case 'p':
           ++pflag;
           break;

      case 'P':
           ++Pflag;
           break;

      case 'q':
           ++qflag;
           break;

      case 'R':
           ++Rflag;
           break;

      case 'r':
           RFileName = optarg;
           if (strpbrk(RFileName,"?*[]"))
              PERROR (("Wildcard not allowed `%s'",RFileName));
           break;

      case 's':
           snaplen = atoi (optarg);
           if (snaplen <= 0)
              PERROR (("invalid snaplen %s", optarg));
           break;

      case 'S':
           ++Sflag;
           break;

      case 't':
           --tflag;
           break;

      case 'T':
           if (!stricmp(optarg, "vat"))
                packettype = PT_VAT;
           else if (!stricmp(optarg, "wb"))
                packettype = PT_WB;
           else if (!stricmp(optarg, "rpc"))
                packettype = PT_RPC;
           else if (!stricmp(optarg, "rtp"))
                packettype = PT_RTP;
           else if (!stricmp(optarg, "rtcp"))
                packettype = PT_RTCP;
           else if (!stricmp(optarg, "snmp"))
                packettype = PT_SNMP;
           else if (!stricmp(optarg, "sunrpcrm"))
                packettype = PT_SUNRPCRM;
           else if (!stricmp(optarg, "w3mux"))
                packettype = PT_W3MUX;
           else if (!stricmp(optarg, "iiop"))
                packettype = PT_IIOP;
           else if (!stricmp(optarg, "corba"))
                packettype = PT_CORBA;
           else if (!stricmp(optarg, "anep"))
                packettype = PT_ANEP;
           else if (!stricmp(optarg, "srm2"))
                packettype = PT_SRM2;
           else if (!stricmp(optarg, "wc"))
                packettype = PT_WC;
           else if (!stricmp(optarg, "cnfp"))
                packettype = PT_CNFP;
           else PERROR (("Legal packet types are:\n  vat, wb, rpc, "
                         "rtp, rtcp, cnfp, snmp, sunrpcrm, w3mux, iiop, "
                         "corba, anep, srm2 or wc"));
           break;

      case 'u':
           ++uflag;
           break;

      case 'v':
           ++vflag;
           break;

      case 'w':
           ++wflag;
           WFileName = optarg;
           use_conio = 0;
           break;

      case 'x':
           ++xflag;
           break;

      case 'X':
           ++xflag;
           ++xaflag;
           break;

      case 'y':
           dlt_name = optarg;
           dlt = pcap_datalink_name_to_val (dlt_name);
           if (dlt < 0)
              PERROR (("invalid data link type %s", dlt_name));
           break;

      case 'A':
           Aflag++;
           xflag++;
           xaflag = 0;
           break;

      case '?':
      case 'h':
           program_usage (1);
           return (0);

      default:
           program_usage (0);
           return (0);
    }
  }

  if (aflag && nflag)
     PERROR (("-a and -n options are incompatible"));

  if (Hflag && RFileName)
     PERROR (("-H and -r options are incompatible"));

  if (dlt_name && RFileName)
     PERROR (("-y and -r options are incompatible"));

  if (Rflag && !RFileName)
     PERROR (("-R option requires -r <file> option"));

  if (fflag && RFileName)
     PERROR (("-f and -r options are incompatible"));

  if (tflag > 0)
     thiszone = gmt2local (0);

  if (!tcpdump_init(path_program("~/tcpdump.ini")))
     return (0);

  if (Dflag)
  {
    char buf[2048];

    pcap_list_devices (buf, sizeof(buf), 1);
    printf ("Supported network devices:\n%s", buf);
    return (0);
  }

  if (RFileName)
  {
    /*
     * We don't need network access, so just read the capture file
     */
    pd = pcap_open_offline (RFileName, ebuf);
    if (!pd)
       PERROR ((ebuf));

    dlt = pcap_datalink(pd);
    printf ("reading from file %s, link-type %u (", RFileName, dlt);
    dlt_name = pcap_datalink_val_to_name (dlt);
    printf ("%s)\n", dlt_name ? dlt_name : "*unknown*");

    localnet = 0;
    netmask  = 0;
    if (Rflag)
       pcap_set_wait (pd, NULL, 1);
  }
  else    /* open pcap to capture live network packets */
  {
    int i;

    if (!pcap_device)
    {
      pcap_device = pcap_lookupdev (ebuf);
      if (!pcap_device)
         PERROR ((ebuf));
    }

    pd = pcap_open_live (pcap_device, snaplen, !pflag, 0, ebuf);
    if (!pd)
       PERROR ((ebuf));
    if (!pd)
       WARNING ((ebuf));

    if (Lflag)
       show_dlts_and_exit (pd);

    if (dlt >= 0)
    {
      if (pcap_set_datalink(pd, dlt) < 0)
         PERROR (("%s", pcap_geterr(pd)));
      fprintf (stderr, "%s: data link type %s\n", program_name, dlt_name);
    }

    i = pcap_snapshot (pd);
    if (snaplen < i)
    {
      WARNING (("snaplen raised from %d to %d", snaplen, i));
      snaplen = i;
    }
    if (pcap_lookupnet (pcap_device, (bpf_u_int32*)&localnet,
                        (bpf_u_int32*)&netmask, ebuf) < 0)
    {
      localnet = 0;
      netmask  = 0;
      WARNING ((ebuf));
    }
  }

  /* Hook handlers for SIGINT, SIGBREAK and SIGQUIT
   */
  init_handlers();

  if (infile)
       cmdbuf = read_infile (infile);
  else cmdbuf = copy_argv (&argv[optind]);

  if (pcap_compile (pd, &fcode, cmdbuf, Oflag, netmask) < 0)
  {
    program_end (0);
    PERROR ((pcap_geterr(pd)));
  }

  if (snaplen < 102 && fcode.bf_insns[3].k == 36)
     WARNING (("snaplen < 102, XTP headers will be truncated"));

  if (dflag)
  {
    program_end (1);
    bpf_dump (&fcode, dflag);
    return (0);
  }

  init_addrtoname (nflag, localnet, netmask);

  if (pcap_setfilter(pd,&fcode) < 0)
  {
    program_end (0);
    PERROR ((pcap_geterr(pd)));
  }

  if (WFileName)
  {
    pcap_dumper_t *p = pcap_dump_open (pd, WFileName);

    if (!p)
    {
      program_end (0);
      PERROR ((pcap_geterr(pd)));
    }
    printer = pcap_dump;
    pcap_userdata = (u_char*)p;
    if (vflag)
       pcap_verbose (1);
  }
  else
  {
    printer = lookup_printer (pcap_datalink(pd));
    pcap_userdata = 0;
  }

  if (use_conio)
  {
#ifdef USE_CONIO
    set_conio();
#else
    program_end (1);
    PERROR (("`-C' option not supported in this version."));
#endif
  }

  if (!RFileName)
  {
    char buf[50];
    sprintf (buf, "%s: listening on %s\n",
             program_name, pcap_device ? pcap_device : "all devices");

#ifdef USE_CONIO
    if (use_conio)
         PRINTF (buf);
    else fprintf (stderr, buf);
#else
    fprintf (stderr, buf);
#endif
  }

  if (pcap_loop(pd,count,printer,pcap_userdata) < 0)
  {
    fprintf (stderr, "%s: pcap_loop: %s\n", program_name, pcap_geterr(pd));
    rc = 1;
  }

  dump_stats();
  program_end (1);

  if (Hflag && !RFileName)
     dump_hostnames();

  return (rc);
}

static void program_end (int free_mem)
{
  static int done = 0;

  if (!done)
  {
    pcap_close (pd);
    if (free_mem)
    {
      free (pd);
      pd = NULL;
    }
    done = 1;
  }
}

/*
 * Can't print the summary if reading from a savefile
 */
static void dump_stats (void)
{
  struct pcap_stat    stat;
  struct pcap_stat_ex stat_ex;

  if (!pd || pcap_file(pd))
     return;

  if (pcap_stats(pd,&stat) < 0)
     fprintf (stderr, "pcap_stats: %s\n", pcap_geterr(pd));
  else
  {
    fprintf (stderr, "packets received by filter: %-6lu\n", stat.ps_recv);
    fprintf (stderr, "packets passed thru filter: %-6lu\n", pcap_filter_packets());
     
    if (stat.ps_recv)
    {
      fprintf (stderr, "packets dropped by kernel/pcap: %-6lu (%.1f%%)\n",
               stat.ps_drop,
               100.0 * (double)stat.ps_drop / (double)stat.ps_recv);
      fprintf (stderr, "packets dropped by driver/hw:   %-6lu (%.1f%%)\n",
               stat.ps_ifdrop,
               100.0 * (double)stat.ps_ifdrop / (double)stat.ps_recv);
    }
  }

  if (pcap_stats_ex (pd,&stat_ex) >= 0)    /* extended statistics */
  {
    fprintf (stderr,
      "details: packets  %lu, bytes %lu, dropped %lu, errors %lu, collisions %lu\n"
      "         FIFO err %lu, overruns %lu, framing %lu, len err %lu, CRC err %lu\n",
      stat_ex.rx_packets, stat_ex.rx_bytes, stat_ex.rx_dropped,
      stat_ex.rx_errors, stat_ex.collisions, stat_ex.rx_fifo_errors,
      stat_ex.rx_over_errors, stat_ex.rx_frame_errors,
      stat_ex.rx_length_errors, stat_ex.rx_crc_errors);
  }
}

/* 
 * Like default_print() but data need not be aligned
 */
void default_print_unaligned (const u_char *cp, u_int length)
{
  int   nshorts = length / sizeof(u_short);
  u_int s, i = 0;

  if (xaflag)
  {
    dump_hex_ascii (cp,length);
    return;
  }
  if (Aflag)
  {
    dump_ascii (cp,length);
    return;
  }

  while (--nshorts >= 0)
  {
    if ((i++ % 8) == 0)
       PUTS ("\n\t\t\t");
    s = *cp++;
    PRINTF (" %02x%02x", s, *cp++);
  }
  if (length & 1)
  {
    if ((i % 8) == 0)
       PUTS ("\n\t\t\t");
    PRINTF (" %02x", *cp);
  }
}

void default_print (const u_char *bp, u_int length)
{
  const u_short *sp;
  u_int i;
  int   nshorts;

  if (xaflag)
  {
    dump_hex_ascii (bp,length);
    return;
  }
  if (Aflag)
  {
    dump_ascii (bp,length);
    return;
  }

#if 0
  if (!xflag)
  {
    PUTCHAR ('\n');
    print_data (bp, length);
    return;
  }
#endif

  if ((unsigned)bp & 1)
  {
    default_print_unaligned (bp, length);
    return;
  }
  sp = (u_short *) bp;
  nshorts = (u_int) length / sizeof(u_short);
  i = 0;
  while (--nshorts >= 0)
  {
    if ((i++ % 8) == 0)
       PUTS ("\n\t\t\t");
    PRINTF (" %04x", ntohs (*sp++));
  }
  if (length & 1)
  {
    if ((i % 8) == 0)
       PUTS ("\n\t\t\t");
    PRINTF (" %02x", *(u_char*)sp);
  }
}

static void dump_ascii (const u_char *bp, u_int length)
{
  char  buf [5000];
  char *ptr = buf;

  *ptr++ = '\n';

  while ((int)length-- > 0)
  {
    if (*bp > 0x7E || isprint(*bp))
       *ptr++ = *bp++;
    else
    {
     /* Convert control characters, e.g. 13 -> "\r", 0 -> "^@"
      */
      switch (*bp)
      {
        case '\r':
             strcpy (ptr,"\\r");
             break;
        case '\n':
             strcpy (ptr,"\\n\n");
             ptr++;
             break;
        case '\t':
             strcpy (ptr,"\\t");
             break;
        default:
             ptr[0] = '^';
             ptr[1] = *bp + '@';
      }
      ptr += 2;
      bp++;
    }
  }
  if (ptr > buf+1)
  {
    *ptr = '\0';
    PUTS (buf);
  }
}

static void dump_hex_ascii (const u_char *bp, u_int length)
{
  u_int ofs, j;

  for (ofs = 0; ofs < length; ofs += 16)
  {
    char buf [80];

    PRINTF ("\n\t%04X: ", ofs);
    for (j = 0; j < 16 && j+ofs < length; j++)
        PRINTF ("%02X ", (unsigned)bp[j+ofs]);

    if (j < 16)
       PRINTF ("%*s", 3*(16-j), "");

    for (j = 0; j < 16 && j+ofs < length; j++)
    {
      u_char ch = bp[j+ofs];
      if (ch < ' ')
          ch = '.';
      buf[j] = ch;
    }
    PRINTF ("%.*s", (int)j, buf);
  }
}


/*
 *  Print a usage message, optionally full details
 */
static void program_usage (int details)
{
  char features[80] = { '\0' };

#if defined (__HIGHC__)
  static char *builder = "HighC/Pharlap";

#elif defined (__DJGPP__)
  static char *builder = "GNU-C/djgpp2";

#elif defined (__WATCOMC__)
  static char *builder = "Watcom-C";

#else
  static char *builder = "??";
#endif

  printf ("Version %s (%s). pcap v%d.%d\n"
          "Usage: %s%s [-?adDeEfhLlnNOpPqRStuvxXC] [-c count] [-F file]\n"
          "               [-i interface] [-M type] [-r file] [-s snaplen]\n"
          "               [-E espkey] [m module] [-T type] [-w file] [filter-expression]\n\n",
          TCPDUMP_VERSION, builder, PCAP_VERSION_MAJOR, PCAP_VERSION_MINOR,
          program_path, program_name);

  if (details)
  {
    printf (
          "options: -a    translate network/broadcast addresses to names\n"
          "         -b    parse only link-layer protocol <p>\n"
          "         -c    stop after receiving count packets\n"
          "         -d    dump compiled filter expression and stop\n"
          "         -D    dump info on supported devices\n"
          "         -e    print link-level header on each line\n"
          "         -E    secret for decrypting ESP packets\n"
          "         -f    print foreign ip-addresses numerically\n"
          "         -F    read filter expression from file\n"
          "         -h    write resolved host names to `hosts.add' file\n"
          "         -i    select interface. Use `-i?' for supported drivers\n"
          "         -L    list available data link types and exit\n"
          "         -l    make stdout line buffered\n"
          "         -M    specify message type for SunRPC protocols (-M? for list)\n"
          "         -m    load SMI MIB module definitions from file\n"
          "         -n    print addresses numerically\n"
          "         -N    don't print domain of host names\n"
          "         -O    don't run packet-matching optimiser\n"
          "         -p    don't set network driver in promiscuous mode\n"
          "         -P    print details for some SunRPC protocols\n"
          "         -q    quick output; prints short packet-trace\n"
          "         -r    read packet data from file\n"
          "         -R    print packets read from file at real pace\n"
          "         -s    truncate the packets to snaplen (%d)\n"
          "         -S    print absolute TCP sequence numbers\n"
          "         -t    don't print a timestamp on each dump line\n"
          "         -tt   print timestamp in Unix style\n"
          "         -ttt  print times in time-deltas\n"
          "         -tttt print time and date (MM/DD/YYYY)\n"
          "         -T    force interpretion to specified type (-T? for list)\n"
          "         -u    print undecoded NFS handles\n"
          "         -v    verbose output; prints IP-tos/ttl etc.\n"
          "         -w    write packet data to file (see -r)\n"
          "         -A    print packet contents in ascii (without headers)\n"
          "         -x    print packet contents in hex\n"
          "         -X    print packet contents in hex and ascii\n"
          "         -C    print using direct screen write\n", DEFAULT_SNAPLEN);

#ifdef USE_INET6
     strcat (features, "IPv6, ");
#endif

#ifdef PRINT_NETBIOS_DETAILS
     strcat (features, "NetBIOS details, ");
#endif

#ifdef PRINT_DECNET_DETAILS
     strcat (features, "DEC-Net details, ");
#endif

#ifdef USE_RSVP
     strcat (features, "RSVP decoding, ");
#endif

#ifdef USE_SSL
     strcat (features, "SSL decoding, ");
#endif

#ifdef LIBSMI
     strcat (features, "MIB modules, ");
#endif

    *strrchr (features, ',') = '.';
    printf ("Built-in features: %s\n", features);
  }
  exit (-1);
}

#ifdef __DJGPP__
/*
 * Disable wildcard expansion on command-line
 */
char **__crt0_glob_function (char *arg)  
{
  return (char**)0;
}

static int old_sigquit = -1;
#endif

/*
 * called on a interrupt signal
 */
static void sigint_task (void)
{
#ifdef __DJGPP__
  if (old_sigquit != -1)
     __djgpp_set_sigquit_key (old_sigquit);
#endif

#ifdef __HIGHC__
  _dx_reset_data();
#endif

#ifdef USE_CONIO
  if (use_conio)
     unset_conio();
  else
#endif
    putc ('\n', stderr);

  dump_stats();
  program_end (1);

  if (Hflag && !RFileName)
     dump_hostnames();

  exit (0);
}

/*
 * Signal handler for SIGINT/SIGBREAK (^C/^Break).
 */
static void sig_handler (int sig)
{
  signal (sig, sig_handler);

  if (sig == SIGINT)
     sigint_task();

#ifdef SIGBREAK
  if (sig == SIGBREAK)
     sigint_task();
#endif

#ifdef SIGQUIT
  if (sig == SIGQUIT)
     sigint_task();
#endif

  fprintf (stderr, "Illegal signal %d\n", sig);
}

/*
 * Setup handlers for SIGINT,SIGBREAK and SIGQUIT.
 * SIGSEGV, SIGILL and SIGFPE handlers are already installed
 * in pcap_open_live(). Should maybe reset them here.
 */
static void init_handlers (void)
{
#if defined(__DJGPP__)
  old_sigquit = __djgpp_set_sigquit_key (1);  /* Make-code for ESC */
  set_cbreak (break_mode);
#endif

#ifdef SIGBREAK
  signal (SIGBREAK, sig_handler);
#endif

#ifdef SIGQUIT
  signal (SIGQUIT, sig_handler);
#endif

  signal (SIGINT, sig_handler);

#if defined(__HIGHC__)
  InstallExcHandler (NULL);
#endif
}

